import pickle
from multiprocessing.connection import Listener

from .DjangoLoadBalancer.LoadBalancerFactory import LoadBalancerFactory
from .DjangoLoadBalancer.local_settings import LOAD_BALANCER

if __name__ == "__main__":
    load_balancer = LoadBalancerFactory.create_load_balancer()
    address = (LOAD_BALANCER['ADDRESS']['HOST'], LOAD_BALANCER['ADDRESS']['PORT'])
    listener = Listener(address, authkey=bytes(LOAD_BALANCER['ADDRESS']['AUTHKEY'], 'utf-8'))

    while True:
        connection = listener.accept()
        while True:
            payload = connection.recv()
            if payload == 'close':
                connection.close()
                break
            else:
                query = pickle.loads(payload)
                result = load_balancer.run_query(query)
                connection.send(pickle.dumps(result))



import os

import django
for dirpath, dirnames, files in os.walk('/'):
    if dirpath[len('./'):].count(os.sep) < 1:
        print(files)
        print(f'Found directory: {dirpath}')

os.environ.setdefault("DJANGO_SETTINGS_MODULE", "LoadBalancer.settings")
django.setup()

from django.conf import settings

LOAD_BALANCER = settings.LOAD_BALANCER
DATABASES = settings.DATABASES



import enum
import threading
from queue import Queue


class Database():
    def __init__(self, name, waiters, executors, engine=None, operational_error=None):
        self.name = name
        self.status = DatabaseStatus.RUNNING
        self.queries = Queue()
        self.has_queries = threading.Semaphore(value=0)
        self.waiters = waiters
        self.executors = executors
        self.engine = engine
        self.operational_error = operational_error
        self.info = Queue(1)

    def run_queries(self):
        while self.has_queries.acquire():
            query = self.queries.get()
            self.waiters[query.wait]._executor = self.executors[query.type]
            self.waiters[query.wait].run_query(query, self)


class DatabaseStatus(enum.Enum):
    RUNNING = "RUNNING"
    DOWN = "DOWN"


class DatabaseEngine(enum.Enum):
    POSTGRESQL = "django.db.backends.postgresql"
    MYSQL = "django.db.backends.mysql"
    SQLITE = "django.db.backends.sqlite3"



import inspect
import pickle
from multiprocessing.connection import Client

from django.db import models

from .Query import Query, Wait, QueryType


def send_query_to_listener(query):
    from .local_settings import LOAD_BALANCER
    address = (LOAD_BALANCER['ADDRESS']['HOST'], LOAD_BALANCER['ADDRESS']['PORT'])
    connection = Client(address, authkey=bytes(LOAD_BALANCER['ADDRESS']['AUTHKEY'], 'utf-8'))
    payload = pickle.dumps(query)
    connection.send(payload)
    result = pickle.loads(connection.recv())
    connection.send('close')
    connection.close()
    return result


def loadbalance_function(name):
    def loadbalanced_function(self, *args, **kwargs):
        query = Query(Wait.WAIT, QueryType.QUERYSET, method=name, args=args, kwargs=kwargs, model=self.model)
        result = send_query_to_listener(query)
        return result

    return loadbalanced_function


def load_balance(cls):
    for name, method in inspect.getmembers(cls, predicate=inspect.isfunction):
        if not name.startswith('_') and 'alters_data' in dir(method):
            setattr(cls, name, loadbalance_function(name))
    return cls


@load_balance
class LoadBalancerQuerySet(models.QuerySet):
    pass


class BaseLoadBalancerManager(models.Manager):
    def get_queryset(self):
        return LoadBalancerQuerySet(self.model)


class LoadBalancerManager(BaseLoadBalancerManager):
    def __init__(self):
        super(LoadBalancerManager, self).__init__()

    def get_queryset(self):
        query = Query(Wait.DONT_WAIT, QueryType.NO_QUERYSET, method='get_queryset', model=self.model)
        result = send_query_to_listener(query)
        return result


class LoadBalancerModel(models.Model):
    models_manager = models.Manager()
    loadbalancer_base_manager = BaseLoadBalancerManager()
    objects = LoadBalancerManager()

    def save(self, *args, **kwargs):
        query = Query(wait=Wait.WAIT, type=QueryType.NO_QUERYSET, method='save', model=self, args=args, kwargs=kwargs)
        result = send_query_to_listener(query)
        return result

    def delete(self, *args, **kwargs):
        query = Query(wait=Wait.WAIT, type=QueryType.NO_QUERYSET, method='delete', model=self, args=args, kwargs=kwargs)
        result = send_query_to_listener(query)
        return result

    class Meta:
        abstract = True



import sqlite3
import threading
from queue import Queue

import psycopg2

from .BalancingAlgorithms.CUDAlgorithms.MultithreadingAlgorithm import \
    MultithreadingAlgorithm
from .BalancingAlgorithms.RAlgorithms.Interval import Interval
from .BalancingAlgorithms.RAlgorithms.RoundRobin import RoundRobin
from .Database import Database, DatabaseEngine
from .Executors import QuerySetExecutor, NoQuerySetExecutor
from .InfoQueryExecutors import ResponseTimeExecutor, NumberOfConnectionsExecutor
from .LoadBalancer import LoadBalancer
from .Query import QueryType, Wait
from .Waiters import WaitWaiter, DontWaitWaiter
from .local_settings import LOAD_BALANCER,DATABASES


class LoadBalancerFactory:
    @classmethod
    def create_load_balancer(cls):
        result = Queue()
        threads, databases = cls._create_databases(result)
        cud_algorithm = cls._create_cud_algorithm(databases, threads)
        r_algorithm = cls._create_r_algorithm(databases)
        return LoadBalancer(result, databases, cud_algorithm, r_algorithm)

    @classmethod
    def _create_databases(cls, result):
        if LOAD_BALANCER['CUD_ALGORITHM'] == "MULTITHREADING":
            threads = []
            databases = []
            for i, name in enumerate(LOAD_BALANCER['DATABASES']):
                database = Database(name, {Wait.WAIT: WaitWaiter(LOAD_BALANCER['WAIT_TIME']),
                                           Wait.DONT_WAIT: DontWaitWaiter()},
                                    {QueryType.QUERYSET: QuerySetExecutor(result),
                                     QueryType.NO_QUERYSET: NoQuerySetExecutor(result)}, DATABASES[name]['ENGINE'],
                                    cls.OperationalErrorGenerator.generate_operational_error(DATABASES[name]['ENGINE']))
                databases.append(database)
                threads.append(threading.Thread(target=database.run_queries, args=(), daemon=True))
                threads[i].start()
            return threads, databases

    @staticmethod
    def _create_cud_algorithm(databases, threads=None):
        if LOAD_BALANCER['CUD_ALGORITHM'] == "MULTITHREADING":
            return MultithreadingAlgorithm(threads, databases)

    @staticmethod
    def _create_r_algorithm(databases):
        if LOAD_BALANCER['R_ALGORITHM']['NAME'] == "ROUND_ROBIN":
            return RoundRobin(databases)
        elif LOAD_BALANCER['R_ALGORITHM']['NAME'] == "INTERVAL_TIME":
            algorithm = Interval(databases)
            for database in databases:
                database.executors[QueryType.INFO_RESPONSE_TIME] = ResponseTimeExecutor()
            threading.Thread(target=algorithm.update_info,
                             args=(LOAD_BALANCER['R_ALGORITHM']['INTERVAL'], QueryType.INFO_RESPONSE_TIME),
                             daemon=True).start()
            return algorithm
        elif LOAD_BALANCER['R_ALGORITHM']['NAME'] == "INTERVAL_NUMBER_OF_CONNECTIONS":
            algorithm = Interval(databases)
            for database in databases:
                database.executors[QueryType.INFO_NUMBER_OF_CONNECTIONS] = NumberOfConnectionsExecutor()
            threading.Thread(target=algorithm.update_info,
                             args=(LOAD_BALANCER['R_ALGORITHM']['INTERVAL'], QueryType.INFO_NUMBER_OF_CONNECTIONS),
                             daemon=True).start()
            return algorithm

    class OperationalErrorGenerator:
        @staticmethod
        def generate_operational_error(engine):
            if engine == DatabaseEngine.POSTGRESQL.value:
                return psycopg2.OperationalError
            elif engine == DatabaseEngine.SQLITE.value:
                return sqlite3.OperationalError



import time

import django
import django.db
from django.db import connections

from .Database import DatabaseStatus


class Waiter:
    def __init__(self, executor=None):
        self._executor = executor

    def run_query(self, query, database):
        pass


class WaitWaiter(Waiter):
    def __init__(self, wait_time, executor=None):
        super(WaitWaiter, self).__init__(executor)
        self._wait_time = wait_time

    def run_query(self, query, database):
        try:
            self._executor.run_query(query, database)
        except django.db.utils.OperationalError:
            database.status = DatabaseStatus.DOWN
            self.wait_for_connection(query, database)

    def wait_for_connection(self, query, database):
        while True:
            time.sleep(self._wait_time)
            try:
                connections[database.name].connect()
            except database.operational_error:
                pass
            else:
                self._executor.run_query(query, database)
                database.status = DatabaseStatus.RUNNING
                break


class DontWaitWaiter(Waiter):
    def run_query(self, query, database):
        try:
            self._executor.run_query(query, database)
        except django.db.utils.OperationalError:
            database.status = DatabaseStatus.DOWN
            raise django.db.utils.OperationalError



from time import sleep

from ..RAlgorithm import RAlgorithm
from ...Database import DatabaseStatus
from ...Query import Query, Wait


class Interval(RAlgorithm):
    def __init__(self, databases):
        self._databases = databases
        self._next_db = 0
        self._info_per_database = {x: 0 for x in range(len(databases))}

    def execute_r(self, query):
        info_per_database = sorted(self._info_per_database, key=self._info_per_database.get)
        while self._databases[info_per_database[self._next_db]].status == DatabaseStatus.DOWN or not self._databases[info_per_database[self._next_db]].queries.empty:
            self._next_db += 1
            if self._next_db == len(self._databases): self._next_db = 0
        db_for_read = self._databases[info_per_database[self._next_db]]
        db_for_read.queries.put(query)
        db_for_read.has_queries.release()
        self._next_db = 0

    def update_info(self, interval, query_type):
        while True:
            print(self._info_per_database)
            for i, db in enumerate(self._databases):
                if not db.info.empty():
                    self._info_per_database[i] = db.info.get()
                db.queries.put(Query(wait=Wait.WAIT, type=query_type))
                db.has_queries.release()
            sleep(interval)
